/*
 * Copyright (c) 2015 Tricoire Sebastien 3dsman@free.fr
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 */

#ifndef OE_LINESTITCH_H
#define OE_LINESTITCH_H

#include "curves/OE_curve.h"
#include "stitchs/OE_stitchs.h"
#include "curves/OE_joincurve.h"
#include "OE_pattern.h"

namespace Pakal { class  Archive; }
	
class OE_linestitch : public OE_stitchs
{
	public:
        OE_linestitch();
		OE_linestitch(OE_thread* thread, float len, float width, OE_pattern* pattern);
		OE_linestitch(OE_thread* thread, float len, float width, OE_pattern* pattern, float offset);
		~OE_linestitch();
		
		virtual void persist(Pakal::Archive* archive) override;

		virtual bool check();

        virtual void refresh();

        virtual void refresh(float dpi);
		
        void setLen(float len);
        void setWidth(float width);
		void setPattern(OE_pattern* stitch);
        void setOffset(float offset);
        void setReverse(bool reverse);

        OE_joincurve* getJoincurve();
        float getLen();
        float getWidth();
		OE_pattern* getPattern();
        float getOffset();
        bool getReverse();
		
		static bool initPatterns();
		
		void delDependency(OE_base * object);
        virtual void refreshDependency();

	protected:
        OE_joincurve* curve = nullptr;
		bool reverse = false;
		float len=1;
		float width=1;
		OE_pattern* pattern = nullptr;
		float offset = 0;

};

#endif // OE_LINESTITCH_H
